# Special groups
from . import required
from . import poly

# Groups
from . import (
    common,
    crossvalidation,
    filtration,
    interpretation,
    likelihood,
    optimizer,
    positional_encoding,
    prior,
    scaling,
    tf_options,
)

groups = [
    common,
    crossvalidation,
    filtration,
    interpretation,
    likelihood,
    optimizer,
    positional_encoding,
    prior,
    scaling,
    tf_options
]

